function dump(value, indent, seen)
  if type(value) ~= "table" then
	  if type(value) == "string" then
      return string.format('%q', value)
    else
      return tostring(value)
    end
  else
    if type(seen) ~= "table" then
      seen = {}
    elseif seen[value] then
      return "{...}"
    end
    seen[value] = true
    indent = indent or ""
    if next(value) == nil then
      return "{}"
    end
    local str = "{"
    local first = true
    for k,v in pairs(value) do
      if first then
        first = false
      else
        str = str..","
      end
      str = str.."\n"..indent.."  ".."["..dump(k).."] = "..dump(v, indent.."  ")
    end
    str = str.."\n"..indent.."}"
    return str
  end
end

function logdump(value, indent, seen)
	world.logInfo(dump(value, indent, seen))
end


function freeSpace(debugDump)
	local contents = world.containerItems(entity.id()) or {}
	local count = 0
	if debugDump then logdump(contents) end
	for _, item in ipairs(contents) do
		if debugDump then logdump(item) end
		if item then count = count + 1 end
	end
	count = storage.space - count
	if count < 0 then count = 0 end
	if debugDump then logdump(count) end
	return count
end

function init(virtual)
	if virtual then return end
	
	storage.inPos = entity.toAbsolutePosition(entity.configParameter("inputPoint") or {0, 0})
	storage.outPos = entity.toAbsolutePosition(entity.configParameter("outputPoint") or {0, 0})
	
	storage.space = nil
	storage.full = nil
end

function onInteraction(args)
end

-- function onInboundNodeChange(args)
	-- if args.node == 0 and args.level then
		-- logdump(freeSpace(true))
	-- end
-- end

function update(dt)
	if storage.space == nil then -- doesn't work in init()
		storage.space = world.containerSize(entity.id()) or 0
	end
	
	if entity.getInboundNodeLevel(1) or entity.getInboundNodeLevel(2) then -- spill items
		local spillAll = entity.getInboundNodeLevel(1)
		local self = entity.id()
		local items = (world.containerItems(self) or {})
		for i = 1, storage.space do
			local item = items[i]
			if item then
				local itemType = world.itemType(item.name)
				local stack = { generic = true, coin = true, consumable = true, object = true, material = true, thrownitem = true }
				if not spillAll and stack[itemType] then item.count = item.count - 1 end
				-- logdump(i.." - "..item.name.." - "..itemType.." - "..item.count)
				if item.count > 0 then
					local consumed = world.containerConsumeAt(self, i - 1, item.count)
					if consumed then
						world.spawnItem(item.name, storage.outPos, item.count, item.parameters)
						break
					end
				end
			end
		end	
	end
	
	if not entity.getInboundNodeLevel(0) then -- intake is not closed
		local self = entity.id()
		local inPos = storage.inPos
		local outPos = storage.outPos
		local radius = 1.5
		local itemIDs = world.itemDropQuery(inPos, radius)
		for _, itemID in ipairs(itemIDs) do
			local item = world.takeItemDrop(itemID, self)
			if item then
				-- local canFit = world.containerItemsCanFit(self, item)
				-- if canFit then
					-- logdump(canFit)
					-- if canFit > 0 then
						local leftovers = world.containerAddItems(self, item)
						if leftovers and leftovers.count > 0 then
							world.spawnItem(leftovers.name, outPos, leftovers.count, leftovers.parameters)
						end
					-- else
						-- world.spawnItem(item.name, outPos, item.count, item.parameters)
					-- end
				-- else
					-- world.spawnItem(item.name, outPos, item.count, item.parameters)
				-- end
			end
		end
	end
	
	local full = freeSpace() <= 0
	if full ~= storage.full then
		entity.setOutboundNodeLevel(0, full)
		storage.full = full
	end
end
